//
// Project          : Boca Systems Printer Sample Code
// Author           : Michael Hall
// Editor           : Visual Studio 2013
// Language         : C Sharp (C#)
// Version          : 7.0.0.0
// Release Date     : 8 November 2018
// Usage            : This program demonstrates bidi communication between Windows OS and Boca Systems printers.
//
// Boca Systems printers can be installed with or without printer drivers and communication differs substantially.
// This project was developed at Boca Systems to demonstrate bi directional communication with a Boca Systems
// printer with and without the use of printer drivers.  The C# source code demonstrates one method each for serial,
// parallel, Ethernet, WiFi and USB/HID communication at the interface level.  
//
// There are two methods of communication demonstrated when using one of two Boca Systems FGL drivers.
// There is the standard Boca Systems FGL direct bidi driver version 2.0.2.0 which has been used for years by the majority of users.  
// Now there is a new Boca Systems FGL spooler bidi driver with a printer monitor version 3.0.5.0 which is being Beta tested
// for a limited group of users.  With the use of the printer monitor, Windows Operating Systems strutures know as 
// PRINTER_INFO_2 status and PRINTER_INFO_6 dwstatus.  The printer states Ready, Jammed, Out of Stock and Offline can be monitored
// by reading the status values through API calls.
//
// Supplied in driversupport.cs is a method to perform direct bidirectional communication through a standard Boca Systems 
// printer driver in RAW text mode, using Advanced Properties Spool print OFF and print directly to the printer ON.
//
// This sample code has been modified recently to accomodate for a new spooler Bidi Boca Systems driver that includes a printer monitor
// using Advanced Properties Spool print ON and print directly to the printer OFF.  Also with the new driver the Ports
// Property of Enable bidirectional support ON.  When the driver is selected a few property settings are read so that
// a Current_Driver_Type can be set.  See the Boca_Standard and Boca_Spooler definitions below.         //4.0.0.0 and //6.0.0.0
//
// Modifications have also been made to USBHIDSupport to accomodate for 32/64 bit access differences.   //6.0.0.0
//
// The <S1> and <S2> buttons added to FrmBoca will issue either FGL status command to the FGL printer.  These were
// added to simply make it easier to check if the printer is ready, communicating and printing correctly.  When the FGL <S1>
// command is sent to an FGL printer, and the printer is ready and communicating, the response will be an XON.  When
// the FGL <S2> command is sent to an FGL printer, and the printer is ready and communicating, the response will include
// two pieces of information, the ticket counter and the firmware version.
//
// By starting at the list_interfaces routine the code will lead to one of the interface support files such as
// comsupport.cs, ipsupport.cs, lptsupport.cs and USBHIDsupport.cs.  Within these files the four channel functions 
// of Open, Close, Write and Read are demonstrated.  
//
// Older Boca Systems printer have bidirectional parallel ports.  The newer Lemur printer have only unidirectional parallel ports.
//
// Included in the bin folder are some sample pcx, bmp and text files that can be used for testing
//
using System;
using System.IO;
using System.IO.Ports;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Net.NetworkInformation;
using System.Threading;
using System.Runtime.InteropServices;
using System.Diagnostics;
using System.Security.Principal;

namespace boca_sample
{

    public partial class FrmBoca : Form
    {
        //6.1.0.0
        [DllImport("shlwapi.dll", SetLastError = true, EntryPoint = "#437")]
        private static extern bool IsOS(int os);
        
        delegate void DisplayTextCallback(string text);
        delegate void DisplayLightCallback(int e);

        public struct BOCA_HID_PRINTER
        {
            public int Handle;
            public int SerialNumber;
            public int ProductID;
        }

        public string Logo_ID_Number;
        public string Current_Port;
        public string Current_Port_Name;
        public string incoming;
        public char[] incoming_char = new char[100];
        public static int choice_index;
        public static BOCA_HID_PRINTER[] Printer_Table = new BOCA_HID_PRINTER[127];
        public static bool checking_for_data = false;
        public bool recovery = false;
        public static bool wait_status = false;
        public static bool speaking = false;
        public static bool status_found = false;
        public static int write_timeout = USBHID.USBBoca.WRITE_TIMEOUT_DEFAULT;
        public static int read_timeout = USBHID.USBBoca.READ_TIMEOUT_DEFAULT;
        public static string path_type_string = "";
        public Color background_default = Color.LightSteelBlue;
        public Color active_frame_color = Color.SteelBlue;
        public static string connection_type = "";                          //6.0.0.0

        public bool TestFlag = false;
        public static string CurrentOperatingSystem;                        //V3.1.0.0
        public static Int32 Printer_Info_6_dwStatus;                        //4.1.0.0
        public static Int32 Printer_Info_2_Status;                          //4.1.0.0
        public bool monitoring = false;                                     //4.0.0.0
        public static string Current_Driver_Type;                                               //6.0.0.0
        public static string Boca_Standard_Driver1 = "Boca Driver - Direct Monitor Bidi";       //6.0.0.0
        public static string Boca_Standard_Driver2 = "Boca Driver - Direct Standard Bidi";      //6.0.0.0
        public static string Boca_Spooler_Driver1 = "Boca Driver - Spooler Bidi";               //6.0.0.0
        public static string Boca_Spooler_Driver2 = "Boca Driver - Spooler No Bidi";            //6.0.0.0
        public static string Pcl_Boca_Driver = "PCL Boca Systems Driver";                       //6.0.0.0
        public static string Non_Boca_Driver = "Non Boca Systems Driver";                       //6.0.0.0
        public static string Boca_Spooler_Driver3 = "Boca Driver - Network Spooler Monitor";    //6.0.0.0
        public string Current_Baud;                                                             //7.0.0.0
        public string Current_Parity;                                                           //7.0.0.0
        public string Current_Bits;                                                             //7.0.0.0
        public int DRMode = 1;                                                                  //7.0.0.0
        public static string PrinterDriverName = "";                                            //7.0.0.0
        public static IntPtr PrinterDriverDefault = new IntPtr(0);                              //7.0.0.0
        public bool BT_Status = false;                                                          //7.0.0.0
        public static string parent_dir = "";                                                   //7.0.0.0
        public static string current_child_dir = "";                                            //7.0.0.0
        public static string new_child_dir = "";                                                //7.0.0.0
        public static bool Bidi = false;                                                        //7.0.0.0
        public static bool Direct = false;                                                      //7.0.0.0

        //define some of the form variables here so closure can be forced whenever a printer reset is issued
        public FrmText ft = null;                               //Define an instance of the Text Entry form
        public FrmGraphics fg = null;                           //Define an instance of the Graphics form

        public const int PRINTER_STATUS_READY = 0x0;                        //4.0.0.0
        public const int PRINTER_STATUS_PAUSED = 0x1;                       //4.0.0.0
        public const int PRINTER_STATUS_ERROR = 0x2;                        //4.0.0.0
        public const int PRINTER_STATUS_JAM = 0x8;                          //4.0.0.0
        public const int PRINTER_STATUS_OUT = 0x10;                         //4.0.0.0
        public const int PRINTER_STATUS_OFFLINE = 0x80;                     //4.0.0.0
        public const int PRINTER_STATUS_PRINTING = 0x400;                   //4.0.0.0
        public const int OS_ANYSERVER = 29;                                 //6.1.0.0

        //used to play back a pre-recorded wav file when status is read from printer and sent to establish_status()
        [System.Runtime.InteropServices.DllImport("winmm.DLL", EntryPoint = "PlaySound", SetLastError = true, CharSet = CharSet.Unicode, ThrowOnUnmappableChar = true)]
        private static extern bool PlaySound(string szSound, System.IntPtr hMod, PlaySoundFlags flags);

        [System.Flags]
        public enum PlaySoundFlags : int
        {
            SND_SYNC = 0x0000,
            SND_ASYNC = 0x0001,
            SND_NODEFAULT = 0x0002,
            SND_LOOP = 0x0008,
            SND_NOSTOP = 0x0010,
            SND_NOWAIT = 0x00002000,
            SND_FILENAME = 0x00020000,
            SND_RESOURCE = 0x00040004
        }

        public FrmBoca()
        {
            InitializeComponent();
            //establish parent and child directories
            establish_directories();
        }

        private void LstInterface_SelectedIndexChanged(object sender, EventArgs e)
        {

            //4.0.1.0
            MonNo.Checked = true;
            Ready.Text = "";
            Jam.Text = "";
            Out.Text = "";
            Offline.Text = "";
            Ready2.Text = "";
            Jam2.Text = "";
            Out2.Text = "";
            Offline2.Text = "";
            Printing.Text = "";
            Paused.Text = "";
            Error.Text = "";
            CmdProperties.Enabled = false;                          //7.0.0.0

            //The selected index will be 0 or 1
            //if communication through printer driver
            if (LstInterface.SelectedIndex == 0)
            {
                Current_Port_Name = "Printer Driver";
                port_open();
            }
            else            //communication at the interface level
            {
                choice_index = 1;                                   //Indicate select interface
                FrmChoice fc = new FrmChoice(this);                 //create an instance of the choice form
                fc.Show();                                          //display form
            }
        }
        
        private void FrmBoca_Load(object sender, EventArgs e)
        {
            LnkBoca1.Links.Add(0, LnkBoca1.Text.Length, "http://www.bocasystems.com");                          //Display Boca Home page
            LnkBoca2.Links.Add(0, LnkBoca2.Text.Length, "http://www.bocasystems.com/onlinesupportform.php");    //6.4.0.0 Display tech support page
            LnkBoca3.Links.Add(0, LnkBoca3.Text.Length, "http://www.bocasystems.com/documents/fgl46.doc");      //Start download of programmer's guide
            Initial_State();

            //Get Operating system information.
            CurrentOperatingSystem = GetOperatingSystem();           //V3.1.0.0

            Display_Text("Current operating system is " + CurrentOperatingSystem);

            if (IntPtr.Size == 8)                       //@mdh 6.0.0.0
                OSText.Text = CurrentOperatingSystem + " - 64 Bit";  // for 64 bit operating systems
            else
                OSText.Text = CurrentOperatingSystem + " - 32 Bit";  // for 32 bit operating systems

            Version vrs = new Version(Application.ProductVersion);
            Text = Text + " Version " + vrs.Major + "." + vrs.Minor + "." + vrs.Build + "." + vrs.Revision;
        }

        //Determine the PC's Windows operating system
        string GetOperatingSystem()
        {
            //bool test;

            //Operating system info
            OperatingSystem osi = Environment.OSVersion;

            //Version info
            Version vi = osi.Version;

            //MessageBox.Show("Version " + Environment.OSVersion.ToString());
            //test = IsOS(OS_ANYSERVER);
            //MessageBox.Show("Server " + test.ToString());

            //Init return value in case of failure
            string osys = "Undetermined";

            if (osi.Platform == PlatformID.Win32Windows)
            {
                //Check for versions of Windows before Windows NT
                switch (vi.Minor)
                {
                    case 0:
                        osys = "95";
                        break;
                    case 10:
                        if (vi.Revision.ToString() == "2222A")
                            osys = "98SE";
                        else
                            osys = "98";
                        break;
                    case 90:
                        osys = "Me";
                        break;
                    default:
                        break;
                }
            }
            else if (osi.Platform == PlatformID.Win32NT)     //else post NT
            {
                switch (vi.Major)
                {
                    case 3:
                        osys = "NT 3.51";
                        break;
                    case 4:
                        osys = "NT 4.0";
                        break;
                    case 5:
                        if (vi.Minor == 0)
                            osys = "2000";
                        else
                            osys = "XP";
                        break;
                    case 6:                                 
                        if (vi.Minor == 0)
                            osys = "Vista";
                        else if (vi.Minor == 1)
                            osys = "7";
                        else if (vi.Minor == 2)
                        {
                            if (IsOS(OS_ANYSERVER))                                 //6.1.0.0
                                osys = "Server 2012 R2";
                            else
                                osys = "8/10";
                        }
                        else if (vi.Minor == 3)
                        {
                            if (IsOS(OS_ANYSERVER))                                 //6.1.0.0
                                osys = "Server 2012 R2";
                            else
                                osys = "8/10";
                        }
                        else
                            osys = "";
                        break;
                    case 10:
                        osys = "10";
                        break;
                    default:
                        break;
                }
            }

            osys = "Windows " + osys;               //prepend Windows unto string

            //Return OS
            return osys;
        }
        
        //When closing down the application, or when changing interface choices shutdown any open port
        public void port_shutdown()
        {
            //if printing through the driver
            if (driver_flag)
            {
                checking_for_data = false;              //6.0.0.0
                RawPrinterHelper.Close_Printer();
                driver_flag = false;    //once channel to driver closed clear the flag.
            }
            else                        //else communicating at interface level
            {
                //Close HID connection, if needed
                if (USBHID.HIDDevice.boca_hid != null)
                    USBHID_Dissconnect();

                //Close any open serial port
                if (Serial_Port.PortAccess.Boca_Serial != null)
                {
                    if (Serial_Port.PortAccess.Boca_Serial.IsOpen)
                        Serial_Port.PortAccess.Boca_Serial.Close();
                }

                if (EthernetPort.PortAccess.boca_ip != null)
                {
                    if (CmdConnect.Text == "Disconnect")
                        ip_connection();        //call to shutdown 
                }
                //clear this flag to exit "while loop" and terminate any read thread
                checking_for_data = false;
            }
        }

        public void USBHID_Dissconnect()
        {
            //if USB/HID connection established, close it down
            if (USBHID.HIDDevice.boca_hid.getConnectionState())
            {
                //shut down the Read loop in the Read thread
                checking_for_data = false;
                Thread.Sleep(100);

                //Since IO blocked on the final ReadFile, send an <S1> to release the 
                //read function waiting on status
                Transmit("<S1>");

                //with read loop now shut down, close read and write handles and
                //abort read thread to complete disconnect from device 
                USBHID.HIDDevice.boca_hid.disconnectDevice();
            }
        }

        public void Display_Text(string s)
        {
            //The lstresult listbox can display 15 lines of text at the current font
            //of Comic Sans MS, 8.25pt.  Once the 16th line of text is displayed a vertical
            //scroll bar will appear on the right.
            string a, b;
            int max_length = 140;
            int x;

            //add string to the listbox
            if (s.Length < max_length)
                this.lstResult.Items.Add(s);
            else                                //need wrap around for text beyond 100 characters
            {
                //see if a space can be found to break sentence
                x = s.IndexOf(" ", max_length);
                if (x < 0)                      //if no space found after 140th char, try 130th
                    x = s.IndexOf(" ", max_length - 10);
                if (x < 0)
                    x = max_length;                    //if that did not work just chop it.
                a = s.Substring(0, x);
                b = s.Substring(x + 1);
                lstResult.Items.Add(a);
                lstResult.Items.Add(b);

            }

            //vertical scroll bar top index = number of lines in list box - screen size (15)
            int i = lstResult.Items.Count - 15;
            //this code is to put the vertical scrollbar at bottom of listbox screen
            //once it appears.
            if (i > 0)
                lstResult.TopIndex = i;         //set topindex of vertical scroll bar

            //refresh screen to force update
            lstResult.Refresh();

        }

        //this callback function will be used by the serial port read thread 
        //to toggle the lights on the main form
        //
        //  Value   Meaning
        //  =====   ==============
        //   0       CD  - Green
        //   1       CD  - Red
        //   2       CTS - Green
        //   3       CTS - Red
        //   4       DSR - Green
        //   5       DSR - Red
        //   6       DTR - Green
        //   7       DTR - Red
        //   8       RTS - Green
        //   9       RTS - Red
        //  10       TD  - Green
        //  11       TD  - Red
        //  12       RD  - Green
        //  13       RD  - Red
        //
        public void DisplayLight(int indicator)
        {
            IAsyncResult status;
            try
            {
                //If needs to be invoked, invoke it then
                //call the delegate
                if (!lstResult.Disposing)
                {
                    if (this.lstResult.InvokeRequired)
                    {
                        DisplayLightCallback d = new DisplayLightCallback(DisplayLight);
                        status = this.BeginInvoke(d, new object[] { indicator });
                    }
                    else
                    {

                        //in a Boca Systems printer the serial port pin CD is tied high to a 5 volt line perm.
                        //also depending on which interface (9 pin, 25 pin, rj11/12 jack) the CD line might not be connected.
                        //given that the CDlight indicator will simply copy the CTS line, which is used as CD
                        switch (indicator)
                        {
                            case (0):
                                set_cd(Color.LawnGreen);
                                break;
                            case (1):
                                set_cd(Color.Red);
                                break;
                            case (2):
                                set_cd(Color.LawnGreen);
                                set_cts(Color.LawnGreen);
                                break;
                            case (3):
                                set_cts(Color.Red);
                                set_cd(Color.Red);
                                break;
                            case (4):
                                set_dsr(Color.LawnGreen);
                                break;
                            case (5):
                                set_dsr(Color.Red);
                                break;
                            case (6):
                                set_dtr(Color.LawnGreen);
                                break;
                            case (7):
                                set_dtr(Color.Red);
                                break;
                            case (8):
                                set_rts(Color.LawnGreen);
                                break;
                            case (9):
                                set_rts(Color.Red);
                                break;
                            case (10):
                                set_td(Color.LawnGreen);
                                break;
                            case (11):
                                set_td(Color.Red);
                                break;
                            case (12):
                                set_rd(Color.LawnGreen);
                                break;
                            case (13):
                                set_rd(Color.Red);
                                break;

                            default:
                                break;

                        }
                        //MessageBox.Show("Serial port pin change event " + e.EventType.ToString());

                    }
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
            }

        }

        //display the USB printer serial number
        public void Display_SN(string s)
        {
            serial_number.Text = s;
        }

        //clear messages
        private void CmdClear_Click(object sender, EventArgs e)
        {
            Clear_Display();
        }

        //Scan PC for all Human Interface Devices (HIDs)
        private void CmdScan_Click(object sender, EventArgs e)
        {
            scanner();
        }

        //Scan the PC USB ports for all Boca Systems printers in HID mode
        private int scanner()
        {
            int x, sn = -1;
            int i;
            string model = "";
            
            //Call find hid with serial number set to -1.  This will trigger
            //a scan of all Boca Systems HID mode printers currently installed
            //The Printer_Table will be populated by a call to Find_HIDs
            i = USBHID.HIDDevice.Find_HIDs(sn);

            Display_Text(i.ToString() + " Boca Systems HID printers found");
            for (x = 0; x < i; x++)
            {
                switch (Printer_Table[x].ProductID)
                {
                    case 257:                   //Boca Systems FGL44 Product ID = 0x0101
                        model = " Model FGL44";
                        break;
                    case 258:                   //Boca Systems HP44 Product ID = 0x0102
                        model = " Model HP44";
                        break;
                    case 513:                   //Boca Systems FGL46 Product ID = 0x0201
                        model = " Model FGL46";
                        break;
                    case 514:                   //Boca Systems HP46 Product ID = 0x0202
                        model = " Model HP46";
                        break;
                    default:
                        model = " Model Unknown";
                        break;

                }
                Display_Text(model + " with Serial Number " + Printer_Table[x].SerialNumber.ToString());
            }
            return (i);
        }

        private void CmdConnect_Click(object sender, EventArgs e)
        {
            ip_connection();
        }

        private void ip_connection()
        {
            bool status;

            if (CmdConnect.Text == "Connect")
            {
                boca_sample.FrmBoca.checking_for_data = true;

                //@mdh test making this call here to start read thread and set up m_parent and boca_ip
                EthernetPort.PortAccess fp = new EthernetPort.PortAccess(this);

                Display_Text("Attempting to open IP port.");
                status = EthernetPort.PortAccess.open_ip(IP_Address.Text);
                if (status == true)
                {
                    Display_Text("Success - IP port connection completed.");
                    Enable_General(true);               //turn buttons on
                    CmdConnect.Text = "Disconnect";     //switch button to disconnect function
                    printer_ready();
                }
                else
                {
                    Display_Text("Timeout - IP port connection failed.");
                    port_shutdown();                        //shut down port read thread
                    CmdConnect.Text = "Connect";            //switch button to connect function
                    Enable_General(false);                  //turn buttons off
                    printer_not_found();
                }

            }
            else// CmdConnect.Text must equal "Disconnect"
            {

                status = EthernetPort.PortAccess.close_ip();  //close anything that might be open
                if (status == true)
                {
                    CmdConnect.Text = "Connect";            //switch button to connect function
                    Enable_General(false);                  //turn buttons off
                }
                else
                {
                    Display_Text("Unable to close open IP port");
                }
            }

        }


        //Transmit to the printer using strings
        public void Transmit(string command)
        {
            switch (Current_Port)
            {
                case "Driver":
                    RawPrinterHelper.Write_Printer(command);
                    wait_status = false;
                    break;

                case "HID":
                    USBHID.HIDDevice.write_HID(command);
                    wait_status = false;
                    break;

                case "IP":
                    EthernetPort.PortAccess.write_ip(command);
                    wait_status = false;
                    break;

                case "PAR":
                    ParallelPort.PortAccess.write_lpt(Current_Port_Name, command);
                    break;

                case "SER":
                    Serial_Port.PortAccess.write_serial(command);
                    wait_status = false;
                    break;

                default:
                    Current_Port = "";
                    break;
            }
            if (wait_status)
                wait_for_status();

        }

        //transmit to the printer using byte arrays
        public void Transmitb(byte[] outgoing)
        {

            switch (Current_Port)
            {

                case "Driver":
                    RawPrinterHelper.Write_Printerb(outgoing);
                    wait_status = false;

                    break;

                case "HID":
                    USBHID.HIDDevice.write_HIDb(outgoing);
                    wait_status = false;
                    break;

                case "IP":
                    EthernetPort.PortAccess.write_ipb(outgoing);
                    wait_status = false;
                    break;

                case "PAR":
                    ParallelPort.PortAccess.write_lpt_byte(Current_Port_Name, outgoing);
                    break;

                case "SER":
                    Serial_Port.PortAccess.write_serialb(outgoing);
                    wait_status = false;
                    break;

                default:
                    Current_Port = "";
                    break;
            }

            if (wait_status)
                wait_for_status();

        }


        //Check for incoming data
        public void Receive()
        {
            status_found = false;


            //add received bytes to incoming buffer
            switch (Current_Port)
            {

                case "Driver":
                    incoming += RawPrinterHelper.Read_Printer();
                    break;

                case "HID":
                    incoming += USBHID.HIDDevice.read_HID();
                    break;

                case "IP":
                    incoming += EthernetPort.PortAccess.receive_ip();
                    break;

                case "PAR":
                    incoming += ParallelPort.PortAccess.read_lpt(Current_Port_Name);
                    break;

                case "SER":
                    incoming += Serial_Port.PortAccess.read_serial();
                    break;

                default:
                    Current_Port = "";
                    break;
            }

            if (incoming.Length > 0)
            {
                status_found = true;
                Establish_Status();
            }
            else
                //TxtStatus.Text = "No Data Read";
                Display_Text("No Data Read");       //6.0.0.0
        }
        public void Establish_Status()
        {
            int i = 0, x = 0, length = 0;
            string status_response = "", message = "";
            string filename = "";

            incoming_char = incoming.ToCharArray();
            length = incoming.Length;
            if (RfidMode)                           //7.1.0.0
            {
                if (Current_Port == "HID")          //7.1.0.0 Allow multiple reads on HID
                    RfidResponse += incoming;       //7.1.0.0
                else
                    RfidResponse = incoming;        //7.1.0.0
            }
            incoming = "";

            for (i = 0; i < length; i++)
            {
                x = (int)incoming_char[i];

                //Check for common FGL status
                switch (x)
                {
                    case 1:
                        status_response = "Start of Heading";
                        break;

                    case 2:
                        status_response = "Start of Text";
                        break;

                    case 3:
                        status_response = "Paper Jam Path 1";
                        break;

                    case 4:
                        status_response = "Paper Jam Path 2";
                        break;

                    case 5:
                        status_response = "Test Button Ticket ACK";
                        break;

                    case 6:
                        status_response = "Ticket ACK";
                        break;

                    case 7:
                        status_response = "Wrong File Identifier During Update";
                        break;

                    case 8:
                        status_response = "Invalid Checksum";
                        break;

                    case 9:
                        status_response = "Valid Checksum";
                        break;

                    case 10:
                        status_response = "Out of Paper Path 1";
                        break;

                    case 11:
                        status_response = "Out of Paper Path 2";
                        break;

                    case 15:
                        status_response = "Low Paper";
                        break;

                    case 16:
                        status_response = "Out of Tickets";
                        Random random = new Random();
                        int randomNumber = random.Next(0, 2);
                        if (randomNumber == 0)
                            filename = "load.wav";
                        else
                            filename = "out.wav";
                        recovery = true;
                        break;

                    case 17:
                        status_response = "X-On";
                        if (recovery)
                        {
                            recovery = false;
                            filename = "thanks.wav";
                        }
                        break;

                    case 18:
                        status_response = "Power On";
                        break;

                    case 19:
                        status_response = "X-Off";
                        break;

                    case 20:
                        status_response = "Bad Flash Memory";
                        break;

                    case 21:
                        status_response = "Ticket NAK";
                        break;

                    case 24:
                        status_response = "Ticket Jam";
                        filename = "jam.wav";
                        recovery = true;
                        break;

                    case 25:
                        status_response = "Illegal Data";
                        break;

                    case 26:
                        status_response = "Power Up Problem";
                        break;

                    case 27:
                        status_response = "Ticket NAK";
                        break;

                    case 28:
                        status_response = "Downloading Error";
                        break;

                    case 29:
                        status_response = "Cutter Jam";
                        filename = "jam.wav";
                        recovery = true;
                        break;

                    //this data is probably not status data, but a text response
                    default:
                        status_response = "";
                        break;

                }
                //If a printable character, add to outgoing user message
                if ((x >= 32) && (x < 127))
                {
                    message += incoming_char[i].ToString();
                }
                else
                {
                    if (message.Length > 0)
                    {
                        DisplayText(message);
                        message = "";
                    }
                    if (status_response != "")
                        DisplayText(status_response);
                    if (filename != "")
                    {
                        say(filename);
                        filename = "";
                    }
                }   //end switch

            }      //end next i

            //if any residual message exists display it
            if (message.Length > 0)
            {
                DisplayText(message);
                message = "";
            }

        }  //end of establish_status

        private void CmdPing_Click(object sender, EventArgs e)
        {
            PingReply reply;
            int x,y,z;

            Display_Text("");
            Display_Text("Pinging " + IP_Address.Text + " with 32 bytes of data");
            y = 0;
            z = 0;
            for (x = 0; x < 4; x++)
            {
                reply = EthernetPort.PortAccess.pinger(IP_Address.Text);
                if (reply.Status == IPStatus.Success)
                {
                    Display_Text("Reply from " + reply.Address.ToString() + " Bytes = " + reply.Buffer.Length + " Time = " + reply.RoundtripTime + " mms TTL = " + reply.Options.Ttl + " Status = " + reply.Status.ToString());
                    y++;
                }
                else
                {
                    Display_Text("Status:         " + reply.Status.ToString());
                    z++;
                }
            }
            Display_Text("Ping statistics for " + IP_Address.Text);
            Display_Text("Packets: Sent = 4 Received = " + y.ToString() + " Lost = " + z.ToString());
        }
        private void printer_update()
        {
            Txtready.BackColor = System.Drawing.Color.Yellow;
            Txtready.Text = "Updating";
        }
        private void printer_out()
        {
            Txtready.BackColor = System.Drawing.Color.Yellow;
            Txtready.Text = "Out Of Stock";
        }
        private void printer_ticket_jam()
        {
            Txtready.BackColor = System.Drawing.Color.Yellow;
            Txtready.Text = "Ticket Jam";
        }
        private void printer_cutter_jam()
        {
            Txtready.BackColor = System.Drawing.Color.Yellow;
            Txtready.Text = "Cutter Jam";
        }
        private void printer_busy()
        {
            Txtready.BackColor = System.Drawing.Color.Red;
            Txtready.Text = "Printer Busy";
        }
        private void printer_waiting()
        {
            Txtready.BackColor = System.Drawing.Color.Magenta;
            Txtready.Text = "Waiting for Status";
        }
        private void printer_download()
        {
            Txtready.BackColor = System.Drawing.Color.Magenta;
            Txtready.Text = "Downloading";
        }
        public void printer_ready()
        {
            Txtready.BackColor = System.Drawing.Color.LawnGreen;
            Txtready.Text = "Printer Ready";
        }
        private void printer_reset()
        {
            Txtready.BackColor = System.Drawing.Color.Red;
            Txtready.Text = "Printer Reset";
        }
        public void printer_not_found()
        {
            Txtready.BackColor = System.Drawing.Color.Red;
            Txtready.Text = "Printer Not Found";
        }
        public void printer_port_closed()
        {
            Txtready.BackColor = System.Drawing.Color.Red;
            Txtready.Text = "Printer Port Closed";
            LstInterface.Text = "Select Printer Interface";
            FrmParallel.BackColor = background_default;
            FrmSerial.BackColor = background_default;
            FrmUSB.BackColor = background_default;
            FrmEth.BackColor = background_default;
            checking_for_data = false;
        }

        //3.0.0.0
        public void printer_offline()
        {
            Current_Port = "";
            Txtready.BackColor = System.Drawing.Color.Red;
            MonitorBox.Enabled = false;
            Txtready.Text = "Printer Offline";
            TxtStatus.Text = "";                                        //6.0.0.0
            if (MonYes.Checked == true)                                 //4.0.1.0
            {
                Printer_Info_6_dwStatus = PRINTER_STATUS_OFFLINE;
                Printer_Info_2_Status = PRINTER_STATUS_OFFLINE;         //4.1.0.0
                MonNo.Checked = true;
                Monitor_Printer();
            }
            this.Refresh();

        }
        
        //3.0.0.0 let user know the ports are closed since the printer is offline
        public void show_printer_offline()
        {
            Display_Text("Communication Failure - The printer associated with driver " + pd.PrinterSettings.PrinterName + " appears to be offline.");
            Display_Text("Make sure the printer is correctly setup and turned on.  Then select driver again.");
            Current_Port = "";
            printer_offline();
        }
        
        //3.0.0.0 let user know the ports are closed since the printer is offline
        //using callback delegate DisplayText, not Display_Text since called from seperate thread (read thread)
        public void spo()
        {
            DisplayText("");
            DisplayText("Communication Failure - The printer associated with driver " + pd.PrinterSettings.PrinterName + " appears to be offline.");
            //4.0.1.0
            if (MonYes.Checked == true)
            {
                DisplayText("All open channels have been closed and monitoring of Printer_Info_6 dwStatus");
                DisplayText("and monitoring of Printer_Info_2 Status has been terminated as shown below.");
            }
            else
                DisplayText("All open channels have been closed.");
            DisplayText("");
            DisplayText("Make sure the printer is correctly setup and turned on.  Then select driver again.");
            DisplayText("OFFLINE");
        }

        //6.0.0.0
        public void spo_spool()
        {
            DisplayText("");
            DisplayText("Communication Failure - The printer associated with driver " + pd.PrinterSettings.PrinterName + " appears to be offline.");
            DisplayText("To continue, make sure the printer is correctly setup and turned on.");
            Txtready.BackColor = System.Drawing.Color.Red;
            Txtready.Text = "Printer Offline";
            //DisplayText("OFFLINE");
        }

        //6.0.0.0 display HID interface offline and close channels
        public void spo_hid()
        {
            DisplayText("Communication Failure - The HID printer with serial number " + serial_number.Text + " appears to be offline.");
            DisplayText("Make sure the printer is correctly setup and turned on.  Then select the HID interface again.");
            DisplayText("OFFLINE");
            //MessageBox.Show("The Current HID Printer with serial number " + serial_number + " appears to be offline.");
        }


        public void set_cts(System.Drawing.Color col)
        {
            ShpCTS.BackColor = col;
        }
        public void set_cd(System.Drawing.Color col)
        {
            ShpCD.BackColor = col;
        }
        public void set_dsr(System.Drawing.Color col)
        {
            ShpDSR.BackColor = col;
        }
        public void set_dtr(System.Drawing.Color col)
        {
            ShpDTR.BackColor = col;
        }
        public void set_rd(System.Drawing.Color col)
        {
            ShpRD.BackColor = col;
        }
        public void set_rts(System.Drawing.Color col)
        {
            ShpRTS.BackColor = col;
        }
        public void set_td(System.Drawing.Color col)
        {
            ShpTD.BackColor = col;
        }

        //read data from the printer through the currently open interface
        private void CmdRead_Click(object sender, EventArgs e)
        {
             Receive();
        }

        //Establish which printer path to use, only on multipath printer
        private void CmdPath_Click(object sender, EventArgs e)
        {
            choice_index = 51;                                  //Indicate set path
            FrmChoice fc = new FrmChoice(this);                 //create an instance of the choice form
            fc.Show();                                          //display form
        }

        //FGL Test Ticket
        private void CmdFGLTT_Click(object sender, EventArgs e)
        {
            int count = Convert.ToInt32(number_of_tickets.Text);        //4.0.0.0

            standard(count);                                            //4.0.0.0 print number_of_tickets standard test ticket
        }

        //PCL Test Ticket - print prn file previously output by a driver
        private void CmdPCLTT_Click(object sender, EventArgs e)
        {
            int count = Convert.ToInt32(number_of_tickets.Text);        //4.0.0.0

            download_binary_file("pcl_id.prn", count);                  //4.0.0.0 print number_of_tickets PCL test ticket
        }

        //Send Text String
        private void CmdText_Click(object sender, EventArgs e)
        {
            //FrmText ft = new FrmText(this);                         //create an instance of the text entry form

            choice_index = 2;                                   //Indicate Send Text
            if (ft == null)
            {
                ft = new FrmText(this);
                ft.Show();
            }
            else
                ft.BringToFront();
        }

        //send ascii value
        private void CmdValue_Click(object sender, EventArgs e)
        {
            choice_index = 3;                                   //Indicate Send ASCII Value
            FrmNum fn = new FrmNum(this);                       //create an instance of the number form
            fn.Show();                                          //display form
        }

        //FGL Form Feed
        private void CmdFGLFF_Click(object sender, EventArgs e)
        {
            string fgl_ff = "<p>";                              //FGL Form Feed
            choice_index = 4;                                   //Indicate Send FGL Form Feed
            Transmit(fgl_ff);
            if (wait_status)
                wait_for_status();
        }

        //PCL Form Feed
        private void CmdPCLFF_Click(object sender, EventArgs e)
        {
            char ff = (char)12;                                 //PCL Form Feed
            string pcl_ff = "" + ff;
            choice_index = 5;                                   //Indicate Send PCL Form Feed
            Transmit(pcl_ff);
            if (wait_status)
                wait_for_status();

        }

        //Print Graphics files *.pcx or *.bmp.  Also can download graphics files as a logo
        //into flash to print later
        private void CmdGraph_Click(object sender, EventArgs e)
        {
            choice_index = 8;                                   //Indicate graphics screen
            //FrmGraphics fg = new FrmGraphics(this);
            if (fg == null)
            {
                fg = new FrmGraphics(this);
                fg.Show();
            }
            else
                fg.BringToFront();
        }

        //Download Other Files such as text files, softfont files or driver output files (.prn)
        private void CmdOther_Click(object sender, EventArgs e)
        {                            
            choice_index = 9;                                   //Indicate download other files
            select_file("*", 1);
        }

        //Clear Download Space in the printer
        private void CmdClrDls_Click(object sender, EventArgs e)
        {
            choice_index = 20;                                  //Indicate Clear Download Space
            clear_download_space();
        }

        //Download TTF files
        private void CmdTTF_Click(object sender, EventArgs e)
        {
            choice_index = 21;                                  //Indicate Download TTF file
            select_file("ttf", 1);
        }

        private void CmdPick_Click(object sender, EventArgs e)
        {
            choice_index = 25;                                  //Indicate Pick HID
            FrmNum fn = new FrmNum(this);                       //create an instance of the number form
            fn.Show();                                          //display form

        }

        private void OptYes_CheckedChanged(object sender, EventArgs e)
        {
            Receive();              //read and display any data that may already be in the printer buffer
            wait_status = true;     //now ready to read status automatically on some functions
        }

        private void OptNo_CheckedChanged(object sender, EventArgs e)
        {
            wait_status = false;
        }

        //This routine will attempt to read status from the printer
        //The time period spent waiting for status is controlled by max_count
        public bool wait_for_status()
        {
            bool status = false;
            bool problem = false;
            int x = 0;
            int max_count = 20;

            do
            {
                Receive();
                x++;
                if (x > max_count)
                    problem = true;
                else
                {
                    if (status_found)
                        status = status_found;
                    else
                        Thread.Sleep(100);      //1/10 second pause

                }
            } while ((!status) && (!problem));
            return (status);
        }

        public bool wait_for_xon()
        {
            bool status = true;
            return (status);
        }

        private void CmdStatus_Click(object sender, EventArgs e)
        {
            //This routine checks the HID interface for status.  This is different from 
            //reading the printer status in the printer output buffer.
            int x;

            //There are four possible return values
            //HEX           DEC  Description
            //===========   ===  =========================
            //0x18        = 24   Printer Ready
            //0x20        = 32   Printer Out of Paper
            //0x10        = 16   General Printer Error
            //0xffffffff  - -1   No Response from printer
            x = USBHID.HIDDevice.HID_Status();

            switch (x)
            {

                case 24:
                    Display_Text("Printer Ready");
                    break;


                case 32:
                    Display_Text("Printer Out Of Paper");
                    break;


                case 16:
                    Display_Text("General Printer Error");
                    break;


                case -1:
                    Display_Text("No Response From Printer");
                    break;


                //should never get here                    
                default:
                    Display_Text("USB Status = " + x.ToString());
                    break;
            }   //switch

        }

        private void CmdTime_Click(object sender, EventArgs e)
        {
            choice_index = 13;                                  //Indicate Change HID R/W timeout
            FrmNum fn = new FrmNum(this);                       //create an instance of the number form
            fn.Show();                                          //display form
        }

        //This callback function (delegate) is used by the read thread to return printer
        //status through the establish_status routine.
        private void DisplayText(string text)
        {
            IAsyncResult status;
            try
            {
                //If needs to be invoked, invoke it then
                //call the delegate
                if (!lstResult.Disposing)
                {
                    if (this.lstResult.InvokeRequired)
                    {
                        DisplayTextCallback d = new DisplayTextCallback(DisplayText);
                        status = this.BeginInvoke(d, new object[] { text });
                    }
                    //If already invoked then update label with new value.
                    else
                    {
                        string a, b;
                        int max_length = 120;
                        int x;
                        //set_rd(Color.Red);

                        if (text == "Out of Tickets")
                            printer_out();
                        if ((text == "X-On") && (checking_for_data))
                            printer_ready();
                        if (text == "Ticket Jam")
                            printer_ticket_jam();
                        if (text == "Cutter Jam")
                            printer_cutter_jam();
                        if (text == "OFFLINE")
                        {
                            printer_offline();              //6.0.0.0
                            text = "";
                        }

                        //add string to the listbox
                        if (text.Length < max_length)
                            this.lstResult.Items.Add(text);
                        else                                            //need wrap around for text beyond 100 characters
                        {
                            //see if a space can be found to break sentence
                            x = text.IndexOf(" ", max_length);
                            if (x < 0)                                  //if no space found after 100th char, try 90th
                                x = text.IndexOf(" ", max_length - 10);
                            if (x < 0)
                                x = max_length;                         //if that did not work just chop it.
                            a = text.Substring(0, x);
                            b = text.Substring(x + 1);
                            this.lstResult.Items.Add(a);
                            this.lstResult.Items.Add(b);

                        }

                        //vertical scroll bar top index = number of lines in list box - screen size (12)
                        int i = lstResult.Items.Count - 12;
                        //this code is to put the vertical scrollbar at bottom of listbox screen
                        //once it appears.
                        if (i > 0)
                            lstResult.TopIndex = i;         //set topindex of vertical scroll bar

                        //refresh screen to force update
                        lstResult.Refresh();

                    }
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
            }
        }

 
        //this routine is called from establish status with a wav file name as the parameter
        public void say(string fn)
        {
            if(speaking == true)                    //@mdh 4.1.0.0
                PlaySound(fn, new System.IntPtr(), PlaySoundFlags.SND_SYNC);
        }

        public void standard(int total_ticket_count)
        {
            int j;                          //local loop counter
            string ts1, ts2, ts3, ts4, ts5; //temp strings
            char fs = (char)28;             //file separator
            string port = "";

            //Model standard test ticket            
            ts2 = "<RC52,862>PLACE LETTERS ANYWHERE<RC80,1170><F6><RR>VOID<RC15,1100><F3>TEST TICKET ONLY<RC8,1010><X2><NXL10>*LEMUR 123*<RR><RC78,915>  CODE 39";
            ts3 = "<RU><RC320,740><BS17,30>TICKET & LABEL PRINTER<NR><F3><RC0,300>EXCELLENT PRINT QUALITY<NR><RC30,300> Print any bar code<RC70,300><X2><OL6>^CODE 128^<RC90,250><RR>CODE 128<RC96,340><X3><NXP10>*MONKEY*";
            ts4 = "<RC206,340><X2><NP6>*MONKEY*<F2><HW1,1><NR><F1><SP100,700><LO1><RC371,385><F2><NR>SW1=<VA4> SW2= <VA5> SW3=<VA6><RC371,824>LEFT = 1  RIGHT = 0<RC385,150><F11>Script printing is now standard.";
            //if we are communicating thru a printer driver display the driver port name
            if (Current_Port_Name == "Printer Driver")
            {
                //see if a comma can be found in the port name from the driver
                int x = RawPrinterHelper.port_name.IndexOf(",");
                if (x > 0)
                    port = RawPrinterHelper.port_name.Substring(0, x);
                else
                    port = RawPrinterHelper.port_name;
                ts5 = "<RC430,150><F9>High density printing is clear and readable<RC450,150><F2>High density printing is clear and readable<RC470,150><F1>Legal size printing<RC0,100><LT4><VX1200><NR><F9><RC325,110>SN=<VA1><RC350,110>IP=<VA82><RL><F8><RC320,160>Port=" + port + "<p>";
            }
            else //else display current port name
                ts5 = "<RC430,150><F9>High density printing is clear and readable<RC450,150><F2>High density printing is clear and readable<RC470,150><F1>Legal size printing<RC0,100><LT4><VX1200><NR><F9><RC325,110>SN=<VA1><RC350,110>IP=<VA82><RL><F8><RC320,160>Port=" + Current_Port_Name + "<p>";

            //notify the user as to what is going on
            if (total_ticket_count == 1)
                Display_Text("One Standard FGL Test ticket is now being sent to the printer.");
            else
                Display_Text(total_ticket_count.ToString() + " Standard FGL Test tickets are now being sent to the printer.");

            for (j = 1; j <= total_ticket_count; j++)
            {
                //put the current value of loop counter onto ticket as ticket counter
                ts1 = "<n><DA>" + fs + "<RL><RC320,20><F12>BOCA <F6><RC100,40>" + j.ToString() + "<NR><RC320,110><RL><F6><HW1,1><VA7><RC300,160><F6><RL><HW1,1><RC370,740><RU>GHOSTWRITER<RC48,880><RR><F2>FRIENDLY GHOST LANGUAGE";

                //combine the five lines of text for each ticket and transmit it to the printer
                Transmit(ts1 + ts2 + ts3 + ts4 + ts5);
            }
        }       //end of standard
        
        //This routine aids in the selection of a file from a drive.  A filter is set up based on file type.
        //get filename is called to retrieve the file and download_file is called to send the file to the printer.
        //The file type is passed in as a string.  The ticket count is also passed in to indicate the number of tickets
        //to be printed.  For graphic file types pcx and bmp the ticket count can also be set to zero to indicate
        //the graphic file is to be downloaded into flash memory as a logo, not printed at this time.
        public void select_file(string file_type, int ticket_count)
        {
            string file_name, filter;

            switch (file_type)
            {
                case ("bin"):
                    //filter for only BIN files
                    filter = "bin(*.bin)|*.bin";
                    break;

                case ("fnt"):
                    //filter for only FNT files
                    filter = "fnt(*.fnt)|*.fnt";
                    break;

                case ("prn"):
                    //filter for only prn (printer driver output) files
                    filter = "prn(*.prn)|*.prn";
                    break;

                case ("bmp"):
                    //filter for only BMP files
                    filter = "bmp(*.bmp)|*.bmp";

                    break;

                case ("pcx"):
                    //filter for only PCX files
                    filter = "pcx(*.pcx)|*.pcx";
                    break;

                case ("txt"):
                    //filter for only text files
                    filter = "text(*.txt)|*.txt";
                    break;

                case ("ttf"):
                    //filter for only PCX files
                    filter = "ttf(*.ttf)|*.ttf";
                    break;

                case ("pdf"):                           //7.0.0.0
                    //filter for only text files
                    filter = "pdf(*.pdf)|*.pdf";
                    break;

                default:
                    //no filter
                    filter = "all files(*.*)|*.*";
                    break;
            }

            //open file dialog box to retrieve a file name based on user selection
            file_name = getfilename(filter);
            //if the file_name is not blank, download it
            if (file_name != "")
                download_file(file_name, file_type, ticket_count);

        }

        public string getfilename(string filter)
        {
            //This routine will display the open file dialog box and 
            //return the selected file name
            string file_name = "";            //if no file is selected name is blank

            FileDialog ofd = new OpenFileDialog();

            //open file selection window looking for filtered file types
            //if a filter is passed, use it.
            if (filter != "")
                ofd.Filter = filter;
            if (ofd.ShowDialog() == DialogResult.OK)
            {
                file_name = ofd.FileName;
            }
            ofd.Dispose();
            return (file_name);
        }

        //read/download file routine
        //Depending on the file type, downloading includes FGl commands as prefix and/or suffix
        public void download_file(string file_name, string file_type, int total_ticket_count)
        {
            //All files will be read in byte mode into a byte array and delivered to the printer that way
            //Certain specific file types will need a prefix or suffix string added to the data.
            //Files intended as raw PCX and raw BMP are files of this nature.
            //
            //when downloading a graphic file (bmp,pcx), if ticket count = 0 then the intent is to
            //place the image in flash memory, not print it
            switch (file_type)
            {

                //*.prn => Binary output file from a driver
                case "prn":
                    Display_Text("The Printer Driver output file " + file_name + " is now being sent to the printer.");
                    download_binary_file(file_name, total_ticket_count);
                    break;

                //*.pcx => Binary PCX Graphics file
                case "pcx":
                    Display_Text("The PCX graphics file " + file_name + " is now being sent to the printer.");
                    download_graphic_file(file_name, file_type, total_ticket_count);
                    break;

                //*.bmp => Binary BMP Graphics file
                case "bmp":
                    Display_Text("The BMP graphics file " + file_name + " is now being sent to the printer.");
                    download_graphic_file(file_name, file_type, total_ticket_count);
                    break;

                //*.jpg => Binary JPEG Graphics file  //for future expansion not ready yet for jpeg
                case "jpg":
                    Display_Text("The JPEG graphics file " + file_name + " is now being sent to the printer.");
                    download_graphic_file(file_name, file_type, total_ticket_count);
                    break;

                case "ttf":
                    Display_Text("");
                    Display_Text("The True Type font file " + file_name + " is now being download to the printer.");
                    download_ttf_file(file_name);
                    Display_Text("");
                    Display_Text("Download complete");
                    break;

                //*.PDF
                case "pdf":                 //7.0.0.0
                    Display_Text("The PDF file " + file_name + " is now being sent to the printer.");
                    //Display_Text("Number of copies = " + total_ticket_count);
                    RawPrinterHelper.PDF_Print(file_name);
                    break;

                //by reading text files as bytes one doesn't have to worry about
                //ASCII vs UNICODE, so treat it as a binary file
                //*.txt => Text file
                case "txt":
                    Display_Text("The text file " + file_name + " is now being sent to the printer.");
                    download_binary_file(file_name, total_ticket_count);
                    break;

                //*.* => Text/Binary file needing no prefix or suffix strings
                default:
                    download_binary_file(file_name, total_ticket_count);
                    break;
            }
        }

        //7.0.0.0
        //This routine will cleanup from any configuration change that causes the printer to reset itself
        public void reset()
        {
            //Close any open form except FrmBoca
            //Check FrmText
            if (ft != null)
            {
                ft.Close();
                ft = null;
            }
            //Check FrmGraphics
            if (fg != null)
            {
                fg.Close();
                fg = null;
            }

            if ((Current_Port_Name == "Printer Driver") || (Current_Port_Name == "USB - HID Mode") || (Current_Port_Name == "ETHERNET"))
            {
                reset_flag = true;              //printer forced to reset
                Thread.Sleep(100);              //Wait 1/10 second for the data to completely exit the PC port before shutting down the port
                port_shutdown();                //close any open port
                Display_Text("");
                Display_Text("The printer is now updating itself, so all ports will be closed.  When the printer is ready");
                Display_Text("please select the desired printer again from the drop down menu above.");
                Initial_State();                //return main form to its initial state
                printer_port_closed();          //display printer closed in status
            }
            else
            {
                Display_Text("");
                Display_Text("Please wait for the printer to completely update itself before continuing.");
                Display_Text("");
            }
        }

        //Clear the user download space of all previous downloaded files with Esc c.
        public void clear_download_space()
        {
            //char esc = (char)27;                      //Escape character
            byte escb = (byte)((char)27);               //Escape byte

            byte[] bout = new byte[2];
            bout[0] = escb;             //Esc byte
            bout[1] = 99;           //c
            Transmitb(bout);
            Thread.Sleep(10);      //pause 0.01 seconds
        }

        //read/download text file as a string
        public void download_text_file(string file_name, int total_ticket_count)
        {
            int j;

            string text = File.ReadAllText(file_name);

            for (j = 1; j <= total_ticket_count; j++)
            {
                Transmit(text);
                Thread.Sleep(10);      //pause 0.01 seconds
            }

        }

        //read/download binary file as a byte array
        public void download_binary_file(string file_name, int total_ticket_count)
        {
            int j;

            byte[] bin = File.ReadAllBytes(file_name);
            for (j = 1; j <= total_ticket_count; j++)
            {
                Transmitb(bin);
                Thread.Sleep(10);      //pause 0.01 seconds
            }

        }

        //read/download binary ttf file into printer memory
        public void download_ttf_file(string file_name)
        {
            int i, j, x, y, z;

            //read binary file into byte array
            byte[] bin = File.ReadAllBytes(file_name);

            //establish byte count of file
            x = bin.Length;

            //establish FGL ttf command as a prefix string message, including length
            //of file just read in (byte count).
            string s = "<ttf" + x.ToString() + ">";

            //establish byte count of string
            y = s.Length;

            //construct a byte array for the string
            byte[] bins = new byte[y];

            //put the string into the byte array
            bins = Encoding.ASCII.GetBytes(s);

            //total = file size + prefix string size
            z = x + y;              //z is total

            //construct outgoing byte array as combination of everything
            byte[] bout = new byte[z];

            //first put prefix into outgoing array
            for (j = 0; j < y; j++)
                bout[j] = bins[j];

            //now add file data
            i = 0;
            for (j = y; j < x + y; j++)
            {
                bout[j] = bin[i];
                i++;
            }

            //transmit outgoing array to the printer
            Transmitb(bout);         //send binary ttf data to printer memory

            Thread.Sleep(10);           //pause 0.01 seconds

        }

        //read/download binary bmp or pcx file.  With this routine an image file can be printed
        //or stored in the printer memory for use later.  If the ticket count equals zero then 
        //the intent is to place the image in memory, not print it.  To signal the printer of a 
        //download into memory, an Escape is used first.  To signal end of download an Escape is also last.
        //
        //Example pcx Logo print prefix FGL Command -> <pcx><G1234>Data(1234 bytes)
        //Example pcx Logo memory store prefix FGL Command -> esc<pcx><G1234>Data(1234 bytes)esc
        public void download_graphic_file(string file_name, string file_type, int total_ticket_count)
        {
            int i, j, x, y, z;
            char esc = (char)27;                        //Escape character
            byte escb = (byte)((char)27);               //Escape byte

            //read binary file into byte array
            byte[] bin = System.IO.File.ReadAllBytes(file_name);

            //establish byte count of file
            x = bin.Length;

            //establish FGL graphics command (bmp or pcx) as a prefix string message, including length
            //of file just read in (byte count).
            string s = "<" + file_type + "><G" + x.ToString() + ">";

            //if total_ticket_count = 0 then graphic file is to be downloaded into memory, not printed
            //so prepend ESC to string s
            if (total_ticket_count == 0)
            {
                //4.4.0.0 if logo id number blank then build command with just an escape
                if (Logo_ID_Number == "")
                    s = esc + s;
                else  //4.4.0.0 Download logo using user assigned FGL id number command <ID#> and escape
                {
                    s = "<ID" + Logo_ID_Number + ">" + esc + s;
                    Display_Text("Downloaded into printer memory with user assigned Logo ID set as " + Logo_ID_Number);
                }

            }

            //establish byte count of string
            y = s.Length;

            //construct a byte array for the string
            byte[] bins = new byte[y];

            //put the string into the byte array
            bins = System.Text.Encoding.ASCII.GetBytes(s);

            //total = file size + prefix string size + size of form feed command
            z = x + y + 3;              //z is total

            //construct outgoing byte array to be a combination of everything (FGL commands and data)
            byte[] bout = new byte[z];

            //first put FGL command prefix into outgoing array
            for (j = 0; j < y; j++)
                bout[j] = bins[j];

            //now add file data
            i = 0;
            for (j = y; j < x + y; j++)
            {
                bout[j] = bin[i];
                i++;
            }

            //if count = 0 then graphic file is to be downloaded into memory, not printed
            //so append ESC to outgoing byte array
            if (total_ticket_count == 0)
            {
                bout[j++] = escb;
                //bout[j++] = 32;                 //add space to exit download OK message
                total_ticket_count = 1;         //now set count to one so one copy is downloaded below into memory
            }
            else
            {
                //if printing graphic add FGL form feed command to byte array
                bout[j++] = 60;     //"<"
                bout[j++] = 112;    //"p"
                bout[j++] = 62;     //">"
            }

            //transmit outgoing array to the printer
            for (j = 1; j <= total_ticket_count; j++)
            {
                Transmitb(bout);            //send byte array to the printer
                Thread.Sleep(10);           //pause 0.01 seconds
            }

        }

        //print the text file that will generate a condensed configuration ticket
        //of the printers most popular settings
        public void config_test(int total_ticket_count)
        {
            //Are we to print an FGL or PCL configuration ticket?
            if (FrmBoca.formnum == FrmBoca.FGL)
                download_file("config_fgl.txt", "txt", total_ticket_count);
            else
                download_file("config_hp.txt", "txt", total_ticket_count);
        }

        private void Cmd_Logo_Test_Click(object sender, EventArgs e)
        {
            //This routine is intended to demonstrate "How to download logos in the printer flash".
            //Only monochrome (B&W) pcx or bmp files can be downloaded as logos.  No color & no gray scale.  
            //Below the code will place 5 logos (3 pcx & 2 bmp) into the printer memnory with ID 
            //numbers 1,2,3,10 & 11.  Then five tickets will be printed using the <LD#> FGL command to
            //load each logo file and print it onto a ticket.
            //
            //When pre-loading logos into flash it is necessary to keep track of what is downloaded
            //and in what order.  Logo files placed into flash by default will be assigned an ID number
            //in sequential order.  ID numbers can be assigned by the programmer with the use of
            //the FGL <ID#> command.  By sending the text string command <ID10> just before the download
            //file the file is assigned the ID 10.  The ID number is used by the ticketing software when
            //loading up an image file from memory, to print.
            string file_name;
            string file_type="pcx";

            //To signal download image into memory, set ticket count to zero
            int ticket_count = 0;

            Display_Text("");
            Display_Text("This is a test to demonstrate how to download logos into printer memory to be used later for printing.");
            Display_Text("");

            //Clear Flash - User Download Space, to start with a clean slate 
            Display_Text("First clearing all user download space of previous downloaded logos, symbols and TTFs.");
            Display_Text("");
            clear_download_space();
            Thread.Sleep(100);              //pause 0.1 second

            Display_Text("Downloading images into memory now.");
            Display_Text("");

            //let default ID number be assigned or as shown here use ID 1
            Logo_ID_Number = "1";
            Transmit("<ID1>");
            Display_Text("Downloading file car.pcx with FGL ID = 1");
            file_name = "car.pcx";
            Thread.Sleep(100);              //pause 0.1 second
            download_file(file_name, file_type, ticket_count);
            Thread.Sleep(100);              //pause 0.1 second

            //let default ID number be assigned or as shown here use ID 2
            Logo_ID_Number = "2";
            Transmit("<ID2>");
            Display_Text("Downloading file duck.pcx with FGL ID = 2");
            file_name = "duck.pcx";
            Thread.Sleep(100);              //pause 0.1 second
            download_file(file_name, file_type, ticket_count);
            Thread.Sleep(100);              //pause 0.1 second

            //let default ID number be assigned or as shown here use ID 3
            Logo_ID_Number = "3";
            Transmit("<ID3>");
            Display_Text("Downloading file ghost.pcx with FGL ID = 3");
            file_name = "ghost.pcx";
            Thread.Sleep(100);              //pause 0.1 second
            download_file(file_name, file_type, ticket_count);
            Thread.Sleep(100);              //pause 0.1 second

            //let default ID number be assigned or as shown here use ID 10
            Logo_ID_Number = "10";
            Transmit("<ID10>");
            Thread.Sleep(100);              //pause 0.1 second
            file_type = "bmp";
            file_name = "checkmark.bmp";
            Display_Text("Downloading file checkmark.bmp with FGL ID = 10");
            download_file(file_name, file_type, ticket_count);
            Thread.Sleep(100);              //pause 0.1 second

            //let default ID number be assigned or as shown here use ID 11
            Logo_ID_Number = "11";
            Transmit("<ID11>");
            Thread.Sleep(100);              //pause 0.1 second
            file_name = "lemur_bw.bmp";
            Display_Text("Downloading file lemur_bw.bmp with FGL ID = 11");
            download_file(file_name, file_type, ticket_count);
            Thread.Sleep(100);              //pause 1/10 of second

            Display_Text("");
            Display_Text("Now printing stored logos from memory onto 5 tickets.");
            Display_Text("");
            //generate five tickets with a logo on each.  
            //Use FGL command <LD#> to load logo from memory into print buffer
            //<SP#,#> is used to locate logo on ticket.
            //<RC#,#> is used to locate text on ticket.
            //<F#> is used to select a resident font for text
            Transmit("<RC20,20><F11>Car<SP50,110><LD1><RC400,200><F12>Boca Systems<p>");
            Transmit("<RC20,20><F11>Duck<SP50,110><LD2><RC400,200><F12>Boca Systems<p>");
            Transmit("<RC20,20><F11>Ghost<SP50,110><LD3><RC400,200><F12>Boca Systems<p>");
            Transmit("<RC20,20><F11>Check<SP50,110><LD10><RC400,200><F12>Boca Systems<p>");
            Transmit("<RC20,20><F11>Lemur<SP50,110><LD11><RC400,200><F12>Boca Systems<p>");

        }

        private void LnkBoca1_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            System.Diagnostics.Process.Start(e.Link.LinkData.ToString());
        }

        private void LnkBoca3_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            System.Diagnostics.Process.Start(e.Link.LinkData.ToString());
        }

        private void LnkBoca2_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            System.Diagnostics.Process.Start(e.Link.LinkData.ToString());
        }

        //This routine will read the variable Printer_Info_6_dwStatus and update the 
        //printer_info_6 status fields on FrmBoca
        //The Printer_Info_6_dwStatus veriable is updated in the Is_Printer_Online method
        //4.0.0.0
        public void Monitor_Printer()
        {
            do
            {

                //Check and display each possible status field.  
                //If READY then nothing else needs to be checked
                if (Printer_Info_6_dwStatus == PRINTER_STATUS_READY)
                {
                    Ready.Text = "YES";
                    Jam.Text = "NO";
                    Out.Text = "NO";
                    Offline.Text = "NO";
                    Ready2.Text = "YES";
                    Jam2.Text = "NO";
                    Out2.Text = "NO";
                    Offline2.Text = "NO";
                    //@mdh 4.1.0.0 Printing.Text = "NO";
                    //@mdh 4.1.0.0 Paused.Text = "NO";
                    //@mdh 4.1.0.0 Error.Text = "NO";
                }
                else
                {
                    Ready.Text = "NO";          //Not READY

                    //Check if dwstatus indicates Ticket or Cutter JAM
                    if ((Printer_Info_6_dwStatus & PRINTER_STATUS_JAM) != 0)
                        Jam.Text = "YES";
                    else
                        Jam.Text = "NO";

                    //Check if dwstatus indicates OUT of Ticket Stock
                    if ((Printer_Info_6_dwStatus & PRINTER_STATUS_OUT) != 0)
                        Out.Text = "YES";
                    else
                        Out.Text = "NO";

                    //Check if dwstatus indicates OFFLINE 
                    if ((Printer_Info_6_dwStatus & PRINTER_STATUS_OFFLINE) != 0)
                        Offline.Text = "YES";
                    else
                        Offline.Text = "NO";

                    Ready2.Text = "NO";          //Not READY

                    //Check if dwstatus indicates Ticket or Cutter JAM
                    if ((Printer_Info_2_Status & PRINTER_STATUS_JAM) != 0)
                        Jam2.Text = "YES";
                    else
                        Jam2.Text = "NO";

                    //Check if dwstatus indicates OUT of Ticket Stock
                    if ((Printer_Info_2_Status & PRINTER_STATUS_OUT) != 0)
                        Out2.Text = "YES";
                    else
                        Out2.Text = "NO";

                    //Check if dwstatus indicates OFFLINE 
                    if ((Printer_Info_2_Status & PRINTER_STATUS_OFFLINE) != 0)
                        Offline2.Text = "YES";
                    else
                        Offline2.Text = "NO";

                    //Check if dwstatus indicates PRINTING
                    //@mdh 4.1.0.0 if ((Printer_Info_6_dwStatus & PRINTER_STATUS_PRINTING) != 0)
                        //@mdh 4.1.0.0 Printing.Text = "YES";
                    //@mdh 4.1.0.0 else
                    //@mdh 4.1.0.0 Printing.Text = "NO";

                    //Check if dwstatus indicates PAUSED
                    //@mdh 4.1.0.0 if ((Printer_Info_6_dwStatus & PRINTER_STATUS_PAUSED) != 0)
                        //@mdh 4.1.0.0 Paused.Text = "YES";
                    //@mdh 4.1.0.0 else
                    //@mdh 4.1.0.0 Paused.Text = "NO";

                    //Check if dwstatus indicates General ERROR (Not Jam and Not OUT)
                    //@mdh 4.1.0.0 if ((Printer_Info_6_dwStatus & PRINTER_STATUS_ERROR) != 0)
                        //@mdh 4.1.0.0 Error.Text = "YES";
                    //@mdh 4.1.0.0 else
                    //@mdh 4.1.0.0 Error.Text = "NO";
                }

                //Refresh();
                Ready.Refresh();
                Jam.Refresh();
                Out.Refresh();
                Offline.Refresh();
                Ready2.Refresh();
                Jam2.Refresh();
                Out2.Refresh();
                Offline2.Refresh();
                //@mdh 4.1.0.0 Printing.Refresh();
                //@mdh 4.1.0.0 Paused.Refresh();
                //@mdh 4.1.0.0 Error.Refresh();
                Application.DoEvents();

            } while (monitoring);

        }

        private void MonNo_CheckedChanged(object sender, EventArgs e)
        {
            if (MonNo.Checked)
                monitoring = false;              //4.0.0.0
            CmdFGLFF.Focus();
            Refresh();
        }

        private void MonYes_CheckedChanged(object sender, EventArgs e)
        {
            if (MonYes.Checked)
            {
                monitoring = true;              //4.0.0.0
                Monitor_Printer();
            }
            CmdFGLFF.Focus();
            Refresh();
        }

        //Turn Sound on
        private void OptOn_CheckedChanged(object sender, EventArgs e)
        {
            speaking = true;
        }

        //Turn sound off
        private void OptOff_CheckedChanged(object sender, EventArgs e)
        {
            speaking = false;
        }

        //When an <S1> is transmitted while using bidirectional communication, the printer will
        //respond with an XON (0x11) when ready.  If the printer is not ready nothing will be returned.
        private void CmdS1_Click(object sender, EventArgs e)
        {
            Transmit("<S1>");
        }

        //When an <S2> is transmitted while using bidirectional communication, the printer will
        //respond with a printed ticket count and firmware versio when ready.  It will look 
        //something like 0000160 PROM=FGL46M169, if 160 tickets had printed since last power on. 
        //If the printer is not ready nothing will be returned.
        private void CmdS2_Click(object sender, EventArgs e)
        {
            Transmit("<S2>");
        }

        private void button1_Click(object sender, EventArgs e)
        {
           // RawPrinterHelper.PDF_Print(file_name);
        }

        private void establish_directories()
        {
            char backslash = (char)92;

            //establish current parent and child directory
            current_child_dir = Environment.CurrentDirectory;
            //need to parse existing version number from current_child to make parent path
            //to do this find the last back slash and truncate whatever is to the right of that
            int x = current_child_dir.LastIndexOf(backslash);
            if (x != -1)
                parent_dir = current_child_dir.Substring(0, x);

        }

        //Print a PDF file
        private void CmdPDF_Click(object sender, EventArgs e)
        {
            choice_index = 9;                                   //Indicate download other files
            select_file("pdf", 1);                              //7.0.0.0

        }

        //Change Printer driver properties
        private void CmdProperties_Click(object sender, EventArgs e)
        {
            FrmPref fp = new FrmPref(this);             //7.0.0.0 create an instance of the Driver Prefrences form
            fp.Show();                                  //display form

        }

        //These two BlueTooth functions are used to distinguish BT Serial from standard serial port
        private void BTYes_CheckedChanged(object sender, EventArgs e)
        {
            BT_Status = true;
        }

        private void BTNo_CheckedChanged(object sender, EventArgs e)
        {
            BT_Status = false;
            AutoNo.Checked = true;
        }

        //7.1.0.0 Start of RFID section to write RFID data and then verify same
        //Added RFID test buttons on main panel
        //Default starting block
        string StBlk = "5";
        string RfidResponse = "";
        bool RfidMode = false;

        private void RFWrite_Click(object sender, EventArgs e)
        {
            //Write/Verify Test
            incoming = "";
            RfidMode = true;
            Display_Text("Writing 'Boca' to the RFID Ticket. To verify, will read back and display below.");

            //Write the word "Boca", Read it back and Display
            Transmit("<RFW1," + StBlk + ",0,4>Boca<RFR1," + StBlk + ",4,1>");

            //7.1.0.0 Establish_Status routine puts return value into RfidResponse
            //7.1.0.0 HID interface seems to return multiple byte strings in two or more pieces
            if (Current_Port == "HID")
                Thread.Sleep(2000);             //allow extra time for two or more reads to work on HID
            else
                Thread.Sleep(1000);             //allows time for read routine to work
            if (RfidResponse == "Boca")
                TxtStat.Text = "Pass";
            else
                TxtStat.Text = "Fail";

            RfidMode = false;
            incoming = "";
            RfidResponse = "";
        }

        private void RFRead_Click(object sender, EventArgs e)
        {
            //Read UID/SN using RFID
            Display_Text("The RFID Serial Number is ");
            Transmit("<RFSN2,1>");
        }

        //7.1.0.0 set starting block # based on user input
        private void TxtStBlk_TextChanged(object sender, EventArgs e)
        {
            StBlk = TxtStBlk.Text;
        }

        //7.1.0.0 End of RFID section
        //Added RFID test buttons on main panel

        //8.0.0.0
        //When using this button the messages in the message area can be copied into the system
        //clipboard, and can then be pasted into other editors.  Copy/Paste function.
        private void CmdCopyToClipboard_Click(object sender, EventArgs e)
        {
            try
            {
                StringBuilder sb = new StringBuilder();
                foreach (object row in lstResult.Items)
                {
                    sb.Append(row.ToString());
                    sb.AppendLine();
                }
                sb.Remove(sb.Length - 1, 1); // Just to avoid copying last empty row
                Clipboard.SetData(System.Windows.Forms.DataFormats.Text, sb.ToString());
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }
 
    }
}
